import base64
import copy
import io
import logging
import os.path
import shutil
from typing import Dict, Optional, Tuple, List

import imagesize
from PIL import Image
from image_complete.bmp import is_bmp
from image_complete.jpg import is_jpg
from image_complete.png import is_png
from seppl import MetaDataHandler, LoggingHandler
from ._utils import load_image_from_bytes

FORMAT_JPEG = "JPEG"
FORMAT_PNG = "PNG"
FORMAT_BMP = "BMP"
FORMATS = [
    FORMAT_JPEG,
    FORMAT_PNG,
    FORMAT_BMP,
]
FORMAT_EXTENSIONS = {
    FORMAT_JPEG: ".jpg",
    FORMAT_PNG: ".png",
    FORMAT_BMP: ".bmp",
}


class ImageData(MetaDataHandler, LoggingHandler):

    def __init__(self, source: str = None, image_name: str = None, data: bytes = None,
                 image: Image.Image = None, image_format: str = None, image_size: Tuple[int, int] = None,
                 metadata: Dict = None, annotation=None):
        self._logger = None
        """ for logging. """
        self._source = source
        """ the full path to the image file. """
        self._image_name = image_name
        """ the name of the image file (no path). """
        self._data = data
        """ the binary image data. """
        self._image = image
        """ the Pillow image. """
        self._image_format = image_format
        """ the format of the image. """
        self._image_size = image_size
        """ the size (width, height) tuple of the image. """
        self._metadata = metadata
        """ the dictionary with optional meta-data. """
        self.annotation = annotation
        """ the associated annotation data. """

    def logger(self) -> logging.Logger:
        """
        Returns the logger instance to use.

        :return: the logger
        :rtype: logging.Logger
        """
        if self._logger is None:
            self._logger = logging.getLogger(self.__class__.__name__)
        return self._logger

    @property
    def source(self) -> Optional[str]:
        """
        Returns the source filename.

        :return: the full filename, if available
        :rtype: str
        """
        return self._source

    @property
    def image(self) -> Image.Image:
        """
        Returns the image, loads it if necessary.

        :return: the pillow image data structure, None if not available or failed to load
        :rtype: Image.Image
        """
        if self._image is not None:
            return self._image
        if self._data is not None:
            self._image = load_image_from_bytes(self._data)
            self._image_format = self._image.format
            return self._image
        if self._source is not None:
            self._image_name = os.path.basename(self._source)
            self._image = Image.open(self._source)
            self._image_format = self._image.format
            return self._image
        return None

    @property
    def image_bytes(self):
        """
        Turns the pillow image into bytes using the current image format.

        :return: the generated bytes
        """
        buffer = io.BytesIO()
        if self.image is not None:
            self.image.save(buffer, format=self.image_format)
        return buffer.getvalue()

    @property
    def image_name(self) -> Optional[str]:
        """
        Returns the name of the image.

        :return: the image name, can be None
        :rtype: str
        """
        if self._image_name is not None:
            return self._image_name
        elif self.source is not None:
            return os.path.basename(self.source)
        else:
            return None

    @image_name.setter
    def image_name(self, s: str):
        """
        Sets the new name.

        :param s: the new name
        :type s: str
        """
        self._image_name = s

    @property
    def image_format(self) -> Optional[str]:
        """
        Returns the format of the image.

        :return: the image format, can be None
        :rtype: str
        """
        if self._image_format is None:
            if self.source is not None:
                to_check = self.source
            else:
                to_check = self.data
            if to_check is not None:
                if is_jpg(to_check):
                    self._image_format = FORMAT_JPEG
                elif is_png(to_check):
                    self._image_format = FORMAT_PNG
                elif is_bmp(to_check):
                    self._image_format = FORMAT_BMP
        if self._image_format is None:
            if self.image is None:
                return None
        return self._image_format

    @property
    def image_size(self) -> Optional[Tuple[int, int]]:
        """
        Returns the size tuple (width, height) for the image.

        :return: the width/height tuple, None if failed to determine
        :rtype: tuple
        """
        if self._image_size is not None:
            return self._image_size

        if self._data is not None:
            try:
                self._image_size = imagesize.get(self._data)
                return self._image_size
            except:
                pass

        if self._source is not None:
            try:
                self._image_size = imagesize.get(self._source)
                return self._image_size
            except:
                pass

        if self.image is not None:
            return self.image.size

        return None

    @property
    def image_width(self) -> Optional[int]:
        """
        Returns the width of the image.

        :return: the width, None if failed to determine
        :rtype: int
        """
        size = self.image_size
        if size is None:
            return None
        else:
            return size[0]

    @property
    def image_height(self) -> Optional[int]:
        """
        Returns the height of the image.

        :return: the height, None if failed to determine
        :rtype: int
        """
        size = self.image_size
        if size is None:
            return None
        else:
            return size[1]

    @property
    def data(self) -> bytes:
        """
        Returns the internal data, if any.

        :return: the data
        """
        return self._data

    @data.setter
    def data(self, data: bytes):
        """
        Uses the provided data.

        :param data: the data to use
        """
        self._image_name = self.image_name
        self._source = None
        self._image_format = None
        self._image_size = None
        self._data = data

    def save_image(self, path: str, make_dirs: bool = False) -> bool:
        """
        Saves the image under the specified path.

        :param path: the path to save the image under
        :type path: str
        :param make_dirs: whether to create any missing parent dirs
        :type make_dirs: bool
        :return: whether the file was saved
        :rtype: bool
        """
        if self._source is not None:
            if os.path.exists(path) and os.path.exists(self._source) and os.path.samefile(path, self._source):
                self.logger().warning("Input/output image are the same, skipping!")
                return False
        if make_dirs:
            parent_dir = os.path.dirname(path)
            if not os.path.exists(parent_dir):
                self.logger().info("Creating dir: %s" % parent_dir)
                os.makedirs(parent_dir)
        if (self._data is None) and (self._source is not None) and (os.path.exists(self._source)):
            shutil.copy(self._source, path)
            return True
        if self._image is not None:
            self._image.save(path)
            return True
        if self._data is not None:
            with open(path, "wb") as fp:
                fp.write(self._data)
            return True
        return False

    def has_annotation(self) -> bool:
        """
        Checks whether annotations are present.

        :return: True if annotations present
        :rtype: bool
        """
        return self.annotation is not None

    def has_metadata(self) -> bool:
        """
        Returns whether meta-data is present.

        :return: True if meta-data present
        :rtype: bool
        """
        return self._metadata is not None

    def get_metadata(self) -> Optional[Dict]:
        """
        Returns the meta-data.

        :return: the meta-data, None if not available
        :rtype: dict
        """
        return self._metadata

    def set_metadata(self, metadata: Optional[Dict]):
        """
        Sets the meta-data to use.

        :param metadata: the new meta-data, can be None
        :type metadata: dict
        """
        self._metadata = metadata

    def duplicate(self, source: str = None, force_no_source: bool = None,
                  name: str = None, data: bytes = None,
                  image: Image.Image = None, image_format: str = None,
                  size: Tuple[int, int] = None,
                  metadata: Dict = None, annotation=None):
        """
        Duplicates the container overwriting existing data with any provided data.

        :param source: the source to use
        :type source: str
        :param force_no_source: if True, then source is set to None
        :type force_no_source: bool
        :param name: the name to use
        :type name: str
        :param data: the data to use
        :type data: bytes
        :param image: the Pillow image to use
        :type image: Image.Image
        :param image_format: the image format
        :type image_format: str
        :param size: the size tuple
        :type size: tuple
        :param metadata: the metadata
        :type metadata: dict
        :param annotation: the annotations
        :return: the duplicated container
        """
        if (force_no_source is not None) and force_no_source:
            source = None
        else:
            if source is None:
                source = self._source
        if name is None:
            name = self._image_name
        if (data is None) and (self._data is not None):
            data = copy.deepcopy(self._data)
        # if the source changes, we need to force loading the image
        if ((image is None) and (self._image is not None)) or (source != self._source):
            image = copy.deepcopy(self.image)
        if image_format is None:
            image_format = self._image_format
        if (size is None) and (self._image_size is not None):
            size = copy.deepcopy(self._image_size)
        if (metadata is None) and (self._metadata is not None):
            metadata = copy.deepcopy(self._metadata)
        if (annotation is None) and (self.annotation is not None):
            annotation = copy.deepcopy(self.annotation)

        return type(self)(source=source, image_name=name, data=data,
                          image=image, image_format=image_format, image_size=size,
                          metadata=metadata, annotation=annotation)

    def _annotation_to_dict(self):
        """
        Turns the annotations into a dictionary.

        :return: the generated dictionary
        :rtype: dict
        """
        raise NotImplementedError()

    def to_dict(self, source: bool = True, image: bool = True, annotation: bool = True, metadata: bool = True):
        """
        Returns itself as a dictionary that can be saved as JSON.

        :param source: whether to include the source
        :type source: bool
        :param image: whether to include the image
        :type image: bool
        :param annotation: whether to include the annotations
        :type annotation: bool
        :param metadata: whether to include the metadata
        :type metadata: bool
        :return: the generated dictionary
        :rtype: dict
        """
        result = dict()
        if source and (self.source is not None):
            result["source"] = self.source
        if self.image_name is not None:
            result["name"] = self.image_name
        if self.image_format is not None:
            result["format"] = self.image_format
        if self.image_width is not None:
            result["width"] = self.image_width
        if self.image_height is not None:
            result["height"] = self.image_height
        if image:
            result["image"] = base64.encodebytes(self.image_bytes).decode("ascii")
        if annotation and (self.annotation is not None):
            result["annotation"] = self._annotation_to_dict()
        if metadata and (self.get_metadata() is not None):
            result["metadata"] = copy.deepcopy(self.get_metadata())
        return result


def make_list(data, cls=ImageData) -> List:
    """
    Wraps the data item in a list if not already a list.

    :param data: the data item to wrap if necessary
    :param cls: the type of class to check for
    :return: the list
    :rtype: list
    """
    if isinstance(data, cls):
        data = [data]
    return data


def flatten_list(data: List):
    """
    If the list contains only a single item, then it returns that instead of a list.

    :param data: the list to check
    :type data: list
    :return: the list or single item
    """
    if len(data) == 1:
        return data[0]
    else:
        return data



import argparse
import os
from typing import List

import numpy as np
from PIL import Image
from wai.logging import LOGGING_WARNING

from idc.api import ImageSegmentationData, SplittableStreamWriter, make_list, default_palette, fill_palette, PALETTE_AUTO, PALETTES, AnnotationsOnlyWriter, add_annotations_only_param


class IndexedPngImageSegmentationWriter(SplittableStreamWriter, AnnotationsOnlyWriter):

    def __init__(self, output_dir: str = None,
                 image_path_rel: str = None, palette: str = None, annotations_only: bool = None, background: int = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param output_dir: the output directory to save the image/report in
        :type output_dir: str
        :param image_path_rel: the relative path from the annotations to the images
        :type image_path_rel: str
        :param palette: the palette to use, either a supported palette name (auto|x11|light|dark) or comma-separated list of R,G,B values
        :type palette: str
        :param annotations_only: whether to output only the annotations and not the images
        :type annotations_only: bool
        :param background: the index (0-255) to use as background
        :type background: int
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.image_path_rel = image_path_rel
        self.palette = palette
        self.annotations_only = annotations_only
        self.background = background
        self._palette_list = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-indexed-png-is"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves the annotations as indexed PNG files. The associated JPG images can be placed in folder relative to the annotation."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the images files in. Any defined splits get added beneath there.", required=True)
        parser.add_argument("--image_path_rel", metavar="PATH", type=str, default=None, help="The relative path from the annotations to the images directory", required=False)
        parser.add_argument("-p", "--palette", metavar="PALETTE", type=str, default=PALETTE_AUTO, help="The palette to use; either palette name (%s) or comma-separated list of R,G,B values." % "|".join(PALETTES), required=False)
        parser.add_argument("--background", type=int, help="The index (0-255) to use for the background", required=False, default=0)
        add_annotations_only_param(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.image_path_rel = ns.image_path_rel
        self.palette = ns.palette
        self.annotations_only = ns.annotations_only
        self.background = ns.background

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageSegmentationData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if not os.path.exists(self.output_dir):
            self.logger().info("Creating output dir: %s" % self.output_dir)
            os.makedirs(self.output_dir)
        if self.image_path_rel is None:
            self.image_path_rel = ""
        if self.palette is None:
            self.palette = PALETTE_AUTO
        if self.palette not in PALETTES:
            if "," in self.palette:
                self._palette_list = [int(x) for x in self.palette.split(",")]
                self._palette_list = fill_palette(self._palette_list)
            else:
                raise Exception("Unknown palette: %s" % self.palette)
        else:
            self._palette_list = default_palette(palette=self.palette)
        if self.annotations_only is None:
            self.annotations_only = False

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            sub_dir = self.output_dir
            if self.splitter is not None:
                split = self.splitter.next()
                sub_dir = os.path.join(sub_dir, split)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating sub dir: %s" % sub_dir)
                os.makedirs(sub_dir)

            # image
            path = sub_dir
            if len(self.image_path_rel) > 0:
                path = os.path.join(path, self.image_path_rel)
            os.makedirs(path, exist_ok=True)
            path = os.path.join(path, item.image_name)
            if not self.annotations_only:
                self.logger().info("Writing image to: %s" % path)
                item.save_image(path)

            # annotations
            if item.has_annotation():
                # combine layers
                arr = np.zeros((item.image_height, item.image_width)).astype(dtype=np.uint8)
                for index, label in enumerate(item.annotation.labels, start=1):
                    if label in item.annotation.layers:
                        sub_arr = item.annotation.layers[label]
                        sub_arr = np.where(sub_arr == 255, index, 0).astype(np.uint8)
                        arr += sub_arr
                if self.background > 0:
                    arr = np.where(arr == 0, self.background, arr)
                ann = Image.fromarray(arr, "P")
                ann.putpalette(self._palette_list)
                path = sub_dir
                os.makedirs(path, exist_ok=True)
                path = os.path.join(path, item.image_name)
                path = os.path.splitext(path)[0] + ".png"
                self.logger().info("Writing annotations to: %s" % path)
                ann.save(path)



import argparse
import os
from typing import List

from wai.logging import LOGGING_WARNING
from wai.common.file.report import Report, Field, save
from idc.api import ObjectDetectionData, SplittableStreamWriter, make_list, AnnotationsOnlyWriter, add_annotations_only_param


class AdamsObjectDetectionWriter(SplittableStreamWriter, AnnotationsOnlyWriter):

    def __init__(self, output_dir: str = None, prefix: str = "Object.", annotations_only: bool = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param output_dir: the output directory to save the image/report in
        :type output_dir: str
        :param prefix: the field name prefix to use for storing bbox/polygon definitions in the report files
        :type prefix: str
        :param annotations_only: whether to output only the annotations and not the images
        :type annotations_only: bool
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.prefix = prefix
        self.annotations_only = annotations_only

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-adams-od"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves the bounding box/polygon definitions in an ADAMS .report file alongside the image."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the images/.report files in. Any defined splits get added beneath there.", required=True)
        parser.add_argument("-p", "--prefix", metavar="PREFIX", type=str, default="Object.", help="The field prefix to use in the .report files for identifying bbox/polygon object definitions", required=False)
        add_annotations_only_param(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.prefix = ns.prefix
        self.annotations_only = ns.annotations_only

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if not os.path.exists(self.output_dir):
            self.logger().info("Creating output dir: %s" % self.output_dir)
            os.makedirs(self.output_dir)
        if self.annotations_only is None:
            self.annotations_only = False

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            sub_dir = self.output_dir
            if self.splitter is not None:
                split = self.splitter.next()
                sub_dir = os.path.join(sub_dir, split)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating sub dir: %s" % sub_dir)
                os.makedirs(sub_dir)

            report = Report()
            empty = True
            if item.has_annotation():
                empty = False
                report = item.get_absolute().to_report(prefix=self.prefix)
            if item.has_metadata():
                for k in item.get_metadata():
                    empty = False
                    report.set_value(Field.parse_field(k), item.get_metadata()[k])

            path = os.path.join(sub_dir, item.image_name)
            if not self.annotations_only:
                self.logger().info("Writing image to: %s" % path)
                item.save_image(path)

            if not empty:
                path = os.path.splitext(path)[0] + ".report"
                self.logger().info("Writing report to: %s" % path)
                save(report, path)



import argparse
from typing import List

from seppl.io import Writer, Splitter


def init_splitting_params(writer: Writer, split_names: List[str] = None, split_ratios: List[int] = None):
    """
    Initializes the splitting parameters of the writer.

    :param writer: the writer to initialize
    :type writer: Writer
    :param split_names: the names of the splits, no splitting if None
    :type split_names: list
    :param split_ratios: the integer ratios of the splits (must sum up to 100)
    :type split_ratios: list
    """
    writer.split_names = split_names[:] if (split_names is not None) else None
    writer.split_ratios = split_ratios[:] if (split_ratios is not None) else None
    writer.splitter = None


def add_splitting_params(parser: argparse.ArgumentParser):
    """
    Adds the split ratios/names parameters to the parser.

    :param parser: the parser
    :type parser: argparse.ArgumentParser
    """
    parser.add_argument("-r", "--split_ratios", type=int, default=None, help="The split ratios to use for generating the splits (must sum up to 100)", nargs="+")
    parser.add_argument("-n", "--split_names", type=str, default=None, help="The split names to use for the generated splits.", nargs="+")


def transfer_splitting_params(ns: argparse.Namespace, writer: Writer):
    """
    Transfers the splitting parameters from the parsed namespace into the writer.

    :param ns: the namespace to transfer from
    :type ns: argparse.Namespace
    :param writer: the writer to update
    :type writer: Writer
    """
    writer.split_names = ns.split_names[:] if ((ns.split_names is not None) and (len(ns.split_names) > 0)) else None
    writer.split_ratios = ns.split_ratios[:] if ((ns.split_ratios is not None) and (len(ns.split_ratios) > 0)) else None


def initialize_splitting(writer: Writer):
    """
    Initializes the splitting in the writer.

    :param writer: the writer to initialize, if necessary
    :type writer: Writer
    """
    if not hasattr(writer, "split_names"):
        return
    if (getattr(writer, "split_names") is None) or (getattr(writer, "split_ratios") is None):
        return
    writer.splitter = Splitter(split_ratios=writer.split_ratios, split_names=writer.split_names)
    writer.splitter.initialize()



from typing import List, Tuple
from PIL import ImageColor

# https://en.wikipedia.org/wiki/X11_color_names
X11_COLORS = [
    "#F0F8FF",
    "#FAEBD7",
    "#00FFFF",
    "#7FFFD4",
    "#F0FFFF",
    "#F5F5DC",
    "#FFE4C4",
    "#000000",
    "#FFEBCD",
    "#0000FF",
    "#8A2BE2",
    "#A52A2A",
    "#DEB887",
    "#5F9EA0",
    "#7FFF00",
    "#D2691E",
    "#FF7F50",
    "#6495ED",
    "#FFF8DC",
    "#DC143C",
    "#00FFFF",
    "#00008B",
    "#008B8B",
    "#B8860B",
    "#A9A9A9",
    "#006400",
    "#BDB76B",
    "#8B008B",
    "#556B2F",
    "#FF8C00",
    "#9932CC",
    "#8B0000",
    "#E9967A",
    "#8FBC8F",
    "#483D8B",
    "#2F4F4F",
    "#00CED1",
    "#9400D3",
    "#FF1493",
    "#00BFFF",
    "#696969",
    "#1E90FF",
    "#B22222",
    "#FFFAF0",
    "#228B22",
    "#FF00FF",
    "#DCDCDC",
    "#F8F8FF",
    "#FFD700",
    "#DAA520",
    "#BEBEBE",
    "#808080",
    "#00FF00",
    "#008000",
    "#ADFF2F",
    "#F0FFF0",
    "#FF69B4",
    "#CD5C5C",
    "#4B0082",
    "#FFFFF0",
    "#F0E68C",
    "#E6E6FA",
    "#FFF0F5",
    "#7CFC00",
    "#FFFACD",
    "#ADD8E6",
    "#F08080",
    "#E0FFFF",
    "#FAFAD2",
    "#D3D3D3",
    "#90EE90",
    "#FFB6C1",
    "#FFA07A",
    "#20B2AA",
    "#87CEFA",
    "#778899",
    "#B0C4DE",
    "#FFFFE0",
    "#00FF00",
    "#32CD32",
    "#FAF0E6",
    "#FF00FF",
    "#B03060",
    "#800000",
    "#66CDAA",
    "#0000CD",
    "#BA55D3",
    "#9370DB",
    "#3CB371",
    "#7B68EE",
    "#00FA9A",
    "#48D1CC",
    "#C71585",
    "#191970",
    "#F5FFFA",
    "#FFE4E1",
    "#FFE4B5",
    "#FFDEAD",
    "#000080",
    "#FDF5E6",
    "#808000",
    "#6B8E23",
    "#FFA500",
    "#FF4500",
    "#DA70D6",
    "#EEE8AA",
    "#98FB98",
    "#AFEEEE",
    "#DB7093",
    "#FFEFD5",
    "#FFDAB9",
    "#CD853F",
    "#FFC0CB",
    "#DDA0DD",
    "#B0E0E6",
    "#A020F0",
    "#800080",
    "#663399",
    "#FF0000",
    "#BC8F8F",
    "#4169E1",
    "#8B4513",
    "#FA8072",
    "#F4A460",
    "#2E8B57",
    "#FFF5EE",
    "#A0522D",
    "#C0C0C0",
    "#87CEEB",
    "#6A5ACD",
    "#708090",
    "#FFFAFA",
    "#00FF7F",
    "#4682B4",
    "#D2B48C",
    "#008080",
    "#D8BFD8",
    "#FF6347",
    "#40E0D0",
    "#EE82EE",
    "#F5DEB3",
    "#FFFFFF",
    "#F5F5F5",
    "#FFFF00",
    "#9ACD32",
]

LIGHT_COLORS = [
    "#F0F8FF",
    "#F0FFFF",
    "#F5F5DC",
    "#FFE4C4",
    "#FFEBCD",
    "#FFF8DC",
    "#FFFAF0",
    "#F8F8FF",
    "#F0FFF0",
    "#FFFFF0",
    "#FFF0F5",
    "#E0FFFF",
    "#FFFFE0",
    "#FAF0E6",
    "#F5FFFA",
    "#FDF5E6",
    "#FFF5EE",
    "#FFFAFA",
    "#FFFFFF",
    "#F5F5F5",
]

DARK_COLORS = [
    "#000000",
    "#00008B",
    "#191970",
    "#483D8B",
    "#2F4F4F",
    "#4B0082",
    "#191970",
    "#000080",
]


# Taken from here:
# http://mkweb.bcgsc.ca/colorblind/palettes.mhtml
# http://mkweb.bcgsc.ca/colorblind/palettes/8.color.blindness.palette.txt
COLORBLIND8_COLORS = [
    "#000000",
    "#2271B2",
    "#3DB7E9",
    "#F748A5",
    "#359B73",
    "#d55e00",
    "#e69f00",
    "#f0e442",
]


# Taken from here:
# http://mkweb.bcgsc.ca/colorblind/palettes.mhtml
# http://mkweb.bcgsc.ca/colorblind/palettes/12.color.blindness.palette.txt
COLORBLIND12_COLORS = [
    "#9F0162",
    "#009F81",
    "#FF5AAF",
    "#00FCCF",
    "#8400CD",
    "#008DF9",
    "#00C2F9",
    "#FFB2FD",
    "#A40122",
    "#E20134",
    "#FF6E3A",
    "#FFC33B",
]


# Taken from here:
# http://mkweb.bcgsc.ca/colorblind/palettes.mhtml
# http://mkweb.bcgsc.ca/colorblind/palettes/15.color.blindness.palette.txt
COLORBLIND15_COLORS = [
    "#68023F",
    "#008169",
    "#EF0096",
    "#00DCB5",
    "#FFCFE2",
    "#003C86",
    "#9400E6",
    "#009FFA",
    "#FF71FD",
    "#7CFFFA",
    "#6A0213",
    "#008607",
    "#F60239",
    "#00E307",
    "#FFDC3D",
]


# Taken from here:
# http://mkweb.bcgsc.ca/colorblind/palettes.mhtml
# http://mkweb.bcgsc.ca/colorblind/palettes/24.color.blindness.palette.txt
COLORBLIND24_COLORS = [
    "#003D30",
    "#005745",
    "#00735C",
    "#009175",
    "#00AF8E",
    "#00CBA7",
    "#00EBC1",
    "#86FFDE",
    "#00306F",
    "#00489E",
    "#005FCC",
    "#0079FA",
    "#009FFA",
    "#00C2F9",
    "#00E5F8",
    "#7CFFFA",
    "#004002",
    "#005A01",
    "#007702",
    "#009503",
    "#00B408",
    "#00D302",
    "#00F407",
    "#AFFF2A",
]


PALETTE_AUTO = "auto"
PALETTE_GRAYSCALE = "grayscale"
PALETTE_X11 = "x11"
PALETTE_LIGHT = "light"
PALETTE_DARK = "dark"
PALETTE_COLORBLIND8 = "colorblind8"
PALETTE_COLORBLIND12 = "colorblind12"
PALETTE_COLORBLIND15 = "colorblind15"
PALETTE_COLORBLIND24 = "colorblind24"
PALETTES = [
    PALETTE_AUTO,
    PALETTE_GRAYSCALE,
    PALETTE_X11,
    PALETTE_LIGHT,
    PALETTE_DARK,
    PALETTE_COLORBLIND8,
    PALETTE_COLORBLIND12,
    PALETTE_COLORBLIND15,
    PALETTE_COLORBLIND24,
]


def rgb2yiq(r: int, g: int, b: int) -> float:
    """
    Generates YIQ perceived brightness from RGB colors.

    :param r: the red value
    :type r: int
    :param g: the green value
    :type g: int
    :param b: the blue value
    :type b: int
    :return: the YIQ value
    :rtype: float
    """
    return ((r * 299) + (g * 587) + (b * 114)) / 1000


def default_colors(no_light=True, no_dark=True) -> List:
    """
    Returns a list of default color tuples.

    :param no_light: skips light colors
    :type no_light: bool
    :param no_dark: skips dark colors
    :type no_dark: bool
    :return: the nested list of colors
    :rtype: list
    """
    result = []
    skip = set()
    if no_light:
        for c in LIGHT_COLORS:
            skip.add(c)
    if no_dark:
        for c in DARK_COLORS:
            skip.add(c)
    for c in X11_COLORS:
        if c in skip:
            continue
        result.append(ImageColor.getrgb(c))
    return result


def text_color(color: Tuple[int, int, int], threshold=128) -> Tuple[int, int, int]:
    """
    Computes the text color to use for the given RGB color.

    :param color: the RGB tuple
    :param threshold: the threshold to use
    :return: the text color tuple
    :rtype: tuple
    """
    r, g, b = color
    if rgb2yiq(r, g, b) >= threshold:
        return 0, 0, 0
    else:
        return 255, 255, 255


def create_palette(num_colors: int) -> List[int]:
    """
    Returns a list of palette entries (R,G,B) with the specified number of colors.

    :param num_colors: the number of colors to generate
    :type num_colors: int
    :return: the generated list of colors
    :rtype: list
    """
    return [1 + i // 3 for i in range(3*num_colors)]


def fill_palette(palette: List[int]) -> List[int]:
    """
    Makes sure that there are 256 R,G,B values present. Simply adds grayscale R,G,B values.

    :param palette: the palette to fill up, if necessary
    :type palette: list
    :return: the (potentially) updated list of R,G,B values
    :rtype: list
    """
    if len(palette) < 256*3:
        if len(palette) % 3 != 0:
            raise ValueError("Palette does not contain multiples of three (ie R,G,B values)!")
        palette = palette + create_palette(256 - (len(palette) // 3))
    return palette


def default_palette(palette: str = None) -> List[int]:
    """
    Returns a palette of 255 R,G,B triplets all in a single list, to be used in indexed PNG files.
    Black is always the first color.

    :return: the flat list of R,G,B values
    :rtype: list
    """
    if palette is None:
        palette = PALETTE_AUTO
    if palette not in PALETTES:
        raise ValueError("Unknown palette: %s" % palette)
    if palette == PALETTE_AUTO:
        result = [0, 0, 0,      # black
                  255, 0, 0,    # red
                  0, 255, 0,    # green
                  0, 0, 255,    # blue
                  255, 0, 255,  # magenta
                  255, 255, 0,  # yellow
                  0, 255, 255]  # cyan
    elif palette == PALETTE_GRAYSCALE:
        result = [0, 0, 0]
    elif palette == PALETTE_X11:
        result = [0, 0, 0]
        for c in X11_COLORS:
            if c == "#000000":
                continue
            result.extend(ImageColor.getrgb(c))
    elif palette == PALETTE_LIGHT:
        result = [0, 0, 0]
        for c in LIGHT_COLORS:
            if c == "#000000":
                continue
            result.extend(ImageColor.getrgb(c))
    elif palette == PALETTE_DARK:
        result = [0, 0, 0]
        for c in DARK_COLORS:
            if c == "#000000":
                continue
            result.extend(ImageColor.getrgb(c))
    elif palette == PALETTE_COLORBLIND8:
        result = [0, 0, 0]
        for c in COLORBLIND8_COLORS:
            if c == "#000000":
                continue
            result.extend(ImageColor.getrgb(c))
    elif palette == PALETTE_COLORBLIND12:
        result = [0, 0, 0]
        for c in COLORBLIND12_COLORS:
            if c == "#000000":
                continue
            result.extend(ImageColor.getrgb(c))
    elif palette == PALETTE_COLORBLIND15:
        result = [0, 0, 0]
        for c in COLORBLIND15_COLORS:
            if c == "#000000":
                continue
            result.extend(ImageColor.getrgb(c))
    elif palette == PALETTE_COLORBLIND24:
        result = [0, 0, 0]
        for c in COLORBLIND24_COLORS:
            if c == "#000000":
                continue
            result.extend(ImageColor.getrgb(c))
    else:
        raise Exception("Unhandled palette: %s" % palette)

    result = fill_palette(result)
    return result



import argparse
import os
from typing import List, Iterable, Union

from defusedxml import ElementTree
from wai.logging import LOGGING_WARNING
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from seppl.io import locate_files
from idc.api import ObjectDetectionData, locate_image
from idc.api import Reader


class VOCObjectDetectionReader(Reader):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self._inputs = None
        self._current_input = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-voc-od"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the bounding box from the associated .xml file in PASCAL VOC format."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the XML file(s) to read; glob syntax is supported", required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the XML files to use", required=False, nargs="*")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True, default_glob="*.xml")

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input
        self.logger().info("Reading from: " + str(self.session.current_input))

        xml = ElementTree.parse(self.session.current_input)
        img = os.path.join(
            os.path.dirname(self.session.current_input),
            xml.findtext("folder"),
            xml.findtext("filename")
        )
        if not os.path.exists(img):
            self.logger().warning("Failed to locate image based on information in XML.")
            img = locate_image(self.session.current_input)
            if img is None:
                self.logger().warning("No corresponding image found for: %s" % self.session.current_input)
                self._current_input = None
                yield None

        lobjs = LocatedObjects()
        for obj in xml.findall("object"):
            label = obj.findtext("name")

            # Get the bounding box XML element
            bbox = obj.find("bndbox")

            # Get the boundary co-ordinates
            x_min = int(bbox.findtext("xmin"))
            x_max = int(bbox.findtext("xmax"))
            y_min = int(bbox.findtext("ymin"))
            y_max = int(bbox.findtext("ymax"))

            meta = dict()
            meta["type"] = label
            lobj = LocatedObject(x_min, y_min, x_max - x_min + 1, y_max - y_min + 1, **meta)
            lobjs.append(lobj)

        yield ObjectDetectionData(source=str(img), annotation=lobjs)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0



from ._colors import X11_COLORS, LIGHT_COLORS, DARK_COLORS, COLORBLIND24_COLORS, default_colors, rgb2yiq, text_color
from ._colors import create_palette, fill_palette, default_palette, PALETTE_AUTO, PALETTE_GRAYSCALE, PALETTE_X11, PALETTE_LIGHT, PALETTE_DARK, PALETTE_COLORBLIND8, PALETTE_COLORBLIND12, PALETTE_COLORBLIND15, PALETTE_COLORBLIND24, PALETTES
from ._fonts import DEFAULT_FONT_FAMILY, load_font, text_size
from ._data import ImageData, make_list, flatten_list, FORMATS, FORMAT_JPEG, FORMAT_PNG, FORMAT_BMP, FORMAT_EXTENSIONS
from ._generator import Generator, SingleVariableGenerator
from ._imgcls import ImageClassificationData
from ._imgseg import ImageSegmentationData, ImageSegmentationAnnotations, from_indexedpng, from_bluechannel, from_grayscale, combine_layers, split_layers
from ._objdet import ObjectDetectionData, get_object_label, set_object_label, DEFAULT_LABEL, LABEL_KEY
from ._utils import locate_file, locate_image, load_image_from_bytes, load_image_from_file, load_function
from ._utils import load_labels, save_labels, save_labels_csv
from ._utils import crop_image, pad_image
from ._data_types import DATATYPE_IMGCLS, DATATYPE_OBJDET, DATATYPE_IMGSEG, DATATYPES, data_type_to_class
from ._geometry import locatedobjects_to_shapely, shapely_to_locatedobject, locatedobject_polygon_to_shapely, locatedobject_bbox_to_shapely
from ._geometry import intersect_over_union, COMBINATIONS, INTERSECT, UNION
from ._geometry import merge_polygons, fit_located_object, fit_layers
from ._reader import Reader, parse_reader
from ._filter import parse_filter
from ._writer import parse_writer
from ._writer import BatchWriter, SplittableBatchWriter
from ._writer import StreamWriter, SplittableStreamWriter
from ._writer import AnnotationsOnlyWriter, add_annotations_only_param
from ._splitting import init_splitting_params, add_splitting_params, transfer_splitting_params, initialize_splitting



import argparse
from typing import List, Iterable, Union

from seppl.io import locate_files
from wai.logging import LOGGING_WARNING

from idc.api import DATATYPES, data_type_to_class, ImageData, Reader, load_function


class PythonFunctionReader(Reader):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 function: str = None, data_type: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param function: the function to use (module_name:function_name)
        :type function: str
        :param data_type: the type of output to generate from the images
        :type data_type: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self.function = function
        self.data_type = data_type
        self._inputs = None
        self._current_input = None
        self._output_cls = None
        self._function = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-pyfunc"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the images via the declared function and forwards them as the specified data type. The function must take a string as input and output an iterable of image containers matching the data type."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the image file(s) to read; glob syntax is supported", required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the image files to use", required=False, nargs="*")
        parser.add_argument("-f", "--function", type=str, default=None, help="The Python function to use, format: module_name:function_name", required=True)
        parser.add_argument("-t", "--data_type", choices=DATATYPES, type=str, default=None, help="The type of data to forward", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list
        self.function = ns.function
        self.data_type = ns.data_type

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        if self.data_type is None:
            return [ImageData]
        else:
            return [data_type_to_class(self.data_type)]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.data_type is None:
            raise Exception("No data type defined!")
        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True)
        self._output_cls = data_type_to_class(self.data_type)
        self._function = load_function(self.function)

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input
        self.logger().info("Reading from: " + str(self.session.current_input))
        for item in self._function(self.session.current_input):
            if issubclass(type(item), self._output_cls):
                yield item
            else:
                self.logger().error("Function '%s' did not return an object of type '%s' but of type '%s'!" % (self.function, self._output_cls, type(item)))

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0



import argparse
from typing import List, Iterable, Union

from wai.logging import LOGGING_WARNING
from wai.common.file.report import loadf
from seppl.io import locate_files
from idc.api import ImageClassificationData, locate_image
from idc.api import Reader


class AdamsImageClassificationReader(Reader):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 class_field: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param class_field: the name of the field containing the classification label
        :type class_field: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self.class_field = class_field
        self._inputs = None
        self._current_input = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-adams-ic"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the image classification from the specified class field in the associated .report file."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the report file(s) to read; glob syntax is supported", required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the report files to use", required=False, nargs="*")
        parser.add_argument("-c", "--class_field", metavar="FIELD", type=str, default=None, help="The report field containing the image classification label", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list
        self.class_field = ns.class_field

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ImageClassificationData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.class_field is None:
            raise Exception("No class field defined!")
        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True, default_glob="*.report")

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input
        self.logger().info("Reading from: " + str(self.session.current_input))
        report = loadf(self._current_input)

        meta = dict()
        for field in report:
            meta[field.to_parseable_string()] = report.get_value(field)
        if len(meta) == 0:
            meta = None

        image = locate_image(self._current_input)
        if image is None:
            self.logger().warning("No associated image found: %s" % self._current_input)
            yield None

        if report.has_value(self.class_field):
            yield ImageClassificationData(source=image, annotation=report.get_string_value(self.class_field), metadata=meta)
        else:
            yield ImageClassificationData(source=image, metadata=meta)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0
